/*
 * Created on 22-Mar-2005
 *
 * @author David Balch david.balch@conted.ox.ac.uk
 * 
 */
package uk.ac.ox.tall.wcker;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;

import java.util.LinkedList;

import uk.ac.reload.diva.util.FileUtils;


/**
 * SkeletonExporter is a class that deals with making a skeleton of 
 * files and directories, based on a course hierarchy and sample
 * files. 
 *  
 * @author David Balch david.balch@conted.ox.ac.uk
 *
 */
public class SkeletonExporter extends LinkedList {
	
	boolean debug = WckerSystem.doDebug();
	NeedToKnow thisNtk;
	String operatingDir;
	
	/**
	 * Constructor. Takes a NeedToKnow object and creates a base directory of
	 * the work in-progress directory.
	 * Performs multiple file copies based on the organization structure and 
	 * the file template assigned to each item in the ntn organization.
	 * @param ntk. The NeedToKnow object which will be exported from.
	 */
	public SkeletonExporter(NeedToKnow ntk) {
		if(debug) {
			System.out.println("############## Started SkeletonExporter() ##############");
		}

		// Set variable values
		thisNtk = ntk;
		operatingDir = thisNtk.getOperatingDir() + WckerSystem.fileSeparator();
		// Set export base directory to the work in-progress directory
		String exportBase = operatingDir + "in-progress" + WckerSystem.fileSeparator();
		WckerTree wt = thisNtk.organization;
		
		if(debug) {
			System.out.println("operatingDir: " + operatingDir);
			System.out.println("Export Base: " + exportBase);
		}
		
		//Create the list of files and directories, plus any templates they use
		buildExportList(wt, exportBase);
		
		if(debug) {
			printContents();
		}
	}
	
	/**
	 * Constructor. Takes a NeedToKnow object and a base directory argument.
	 * Performs multiple file copies based on the organization structure and 
	 * the file template assigned to each item in the ntn organization.
	 * @param ntk. The NeedToKnow object which will be exported from.
	 * @param exportBase. Base directory within which the files will be exported to.
	 */
	public SkeletonExporter(NeedToKnow ntk, String exportBase) {
		if(debug) {
			System.out.println("############## Started SkeletonExporter() ##############");
		}

		// Set variable values
		thisNtk = ntk;
		operatingDir = thisNtk.getOperatingDir() + WckerSystem.fileSeparator();
		WckerTree wt = thisNtk.organization;
		
		if(debug) {
			System.out.println("operatingDir: " + operatingDir);
			System.out.println("Export Base: " + exportBase);
		}
		
		//Create the list of files and directories, plus any templates they use
		buildExportList(wt, exportBase);
		
		if(debug) {
			printContents();
		}
	}
	
	/**
	 * Build the list of files and dirctories involved in an export from
	 * the current organization.
	 */
	private void buildExportList(WckerTree wt, String basePath) {
		if(debug) {
			System.out.println("- buildExportList() step -");
		}
		
		boolean isDirectory;
		String thisDestDir;
		String thisDestPath;
		
		//Iterate over the WckerTree, adding each file and directory to the List
		for(int i = 0; i < wt.children.length; ++i) {
			//Isolate the tree node we're dealing with
			WckerTree thisTree = wt.children[i];
			NodeInfo someInfo = new NodeInfo(thisNtk, thisTree);
			
			//Get the values for this node
			OrganisationPayLoad thisTreeInfo = (OrganisationPayLoad)thisTree.payload;
			String thisItem = thisTreeInfo.nodename;
			
			//TODO: the type() method should probably be called getType()
			String thisItemType = someInfo.type(); 
			
			String thisSource = null;
			boolean filter = false; 
			/* TODO: Correctness fix for this workaround currently in use:
			 *   using       (!thisTreeInfo.src.equals(""))
			 *   rather than (thisTreeInfo.src != null)
			 * for both .src and .templatesrc
			 * If no source are template are defined in the .wsi file, 
			 * then .src and .templatesrc should be null
			 */ 
			//if(thisTreeInfo.src != null) {
			if(!thisTreeInfo.src.equals("")) {
				//If there's a source file use it
				thisSource = thisTreeInfo.src;
			//} else if(thisTreeInfo.templatesrc != null) {
			} else if(!thisTreeInfo.templatesrc.equals("")) {
				//If there's no source file use the template
				thisSource = thisTreeInfo.templatesrc;
				//And enable filtering
				filter = true;
			}
			
			/* Build up destination path from:
			 *  - Any preceding path given in basePath
			 *  - A sanitised version of the page name 
			 */
			String thisPathFragment = convertToPath(thisItem);
			String exportType;
			thisDestDir = basePath + thisPathFragment;
			
			if (thisItemType.equals("page")) {
				//If the node type is a page, add it as a page
				thisDestPath = thisDestDir + ".html";
				exportType = "copyFile";
			} else {
				//If the node type is not a page, add it as a directory
				thisDestPath = thisDestDir + WckerSystem.fileSeparator();
				exportType = "addDirectory";
				//Copy relevant support files to this directory
				copySupportFiles(thisDestPath);
			}
			
			//Add the values as an exportList entry
			add(new ExportListEntry(thisItem, exportType, thisDestPath, thisSource, filter));
			if(debug){
				System.out.println("Add ExportListEntry: a '" + thisItemType + "' to " + thisDestPath + ", from " + thisSource);
			}
			
			//Iterate through the tree
			buildExportList(thisTree, thisDestPath);
		}
	}

	/**
	 * Copy support files specified in the wizard.wsi to the specified destination
	 * @param thisDestPath
	 */
	private void copySupportFiles(String thisDestPath) {
		if(thisNtk.supportFiles != null) {
			String thisItem = "Support files";
			String thisSource;
			String exportType = "copyDirectory";
			String[] supportFiles = thisNtk.supportFiles;
			
			for(int i=0; i<supportFiles.length; i++) {
				System.out.println("Add support: " + supportFiles[i]);
				thisSource = supportFiles[i];
				
				add(new ExportListEntry(thisItem, exportType, thisDestPath, thisSource));
			}
		}
	}

	/**
	 * Format a String s so that it is suitable for use as part of a file path
	 * @param theString
	 * @return cleanString
	 */
	private String convertToPath(String theString) {
		//Replace spaces in string with underscores
		String noSpace = theString.replace(" ", "_");
		
		//TODO: Convert string to US-ASCII, so it's safe for all filesystem formats
		String cleanString = noSpace;
		/*
		URLEncoder eu = new URLEncoder();
		try {
			cleanString = eu.encode(theString, "US-ASCII");
		} catch (UnsupportedEncodingException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		*/
		
		return cleanString;
	}

	/**
	 * Export the pages defined in the organization.
	 * @throws IOException
	 */
	public void export() throws IOException {
		if(debug) {
			System.out.println("--- Started SkeletonExporter.export() ---");
			System.out.println("size(): " + size());
			System.out.println("");
		}
		if(size() != 0) {
			//Export each entry in the exportList
			for(int i=0; i < size(); i++) {
				ExportListEntry thisEntry = (ExportListEntry)get(i);
				String thisItem = thisEntry.getItem();
				File thisDest = new File(thisEntry.getDestFile());
				File thisSource = new File(operatingDir + thisEntry.getSourceFile());
				if(debug) {
					System.out.println("Exporting entry " + i + ": " + thisItem);
					System.out.println("Hey, thisSource: "+ thisSource);
				}
				
				if (thisEntry.getType().equals("copyFile")) {
					// If "copyFile", copy the source file to the destination
					if(debug) {
						System.out.println("Copy file:");
						System.out.println("thisSource.exists(): " + thisSource.getAbsolutePath() + " " + thisSource.exists());
						System.out.println("thisDest: " + thisDest.getAbsolutePath());
					}
					FileReader in = new FileReader(thisSource);
					/*
					 * TODO: Implement a string filter for the files being copied here.
					 * This will require the addition of some sort of filter object to 
					 * the ExportListEntry, when present it will be used to substitute
					 * Strings in the file for strings from the organization. 
					 */ 
					FileWriter out = new FileWriter(thisDest);
					int c;
					
					while ((c = in.read()) != -1)
						out.write(c);
					
					in.close();
					out.close();
				} else if (thisEntry.getType().equals("copyDirectory")) {
					// If "copyDirectory", copy a directory and contents to the destination
					if(debug) {
					}
					FileUtils.copyFolder(thisSource.getAbsoluteFile(), thisDest.getAbsoluteFile());
				} else {
					// Otherwise, this is a directory, create it
					if(debug) {
						System.out.println("Make directory: " + thisItem);
					}
					thisDest.mkdir() ;
				}
				if(debug) {
					System.out.println("");
				}
			}
		} else {
			System.out.println("Failed to export files - No exportList");
		}

		if(debug) {
			System.out.println("--- Ended SkeletonExporter.export() ---");
			System.out.println("");
		}
	}

	/*
	 * Debug aid - print out details of all ExportListEntries.
	 */
	private void printContents() {
		System.out.println("######## Skeleton Exporter contents ########");
		System.out.println("size(): " + size());
		for(int i=0; i<size(); i++) {
			System.out.println("ExportListEntry " + i + ": " + ((ExportListEntry)get(i)).getExportDesc());
		}
		System.out.println("###### End Skeleton Exporter contents ######");
	}
}