/*
 * Created on 24-Jan-2005
 *
 */
package uk.ac.ox.tall.wcker;

import java.io.*;
import java.util.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import uk.ac.ox.tall.wcker.WizPackageHandler;
import uk.ac.ox.tall.wcker.WizardCacheItem;

/**
 * WizardCache - Deals with loading, modifying, and saving the 
 * wizard cache file on behalf of {@link ChooseWizard}.
 * <p>
 * The wizard cache file lists {@link WizardCacheItem}s, to
 * avoid having to load, unzip and parse an arbitrary number 
 * of .wwz wizard packages upon each run of WCKER.
 * 
 * @author BalchD
 *
 */
public class WizardCache extends LinkedList{

	private boolean debug = WckerSystem.doDebug();
	private File wizardCacheFile = null;
	private Document cacheDocument = null; 
	
	/**
	 * Turns a file containing wizard names into an
	 * in-memory list of names.
	 *
	 * @param cacheLocation the filepath of the wizard cache XML file
	 */
	
	public WizardCache(String cacheLocation) {
		if (debug) {
			System.out.println("########################################");
			System.out.println("WizardCache class started");
			System.out.println("cacheLocation: " + cacheLocation);
		}
		
		wizardCacheFile = new File(cacheLocation);
		
		//Check if a cachefile exists, if it does - load it, if it doesn't - create one.
		if(wizardCacheFile.exists()){
			if (debug) {
				System.out.println("Loading cache file from: " + cacheLocation);
			}
			//load the cache file
			cacheDocument = LoadXML.loadXml(wizardCacheFile);
			
			//Parse the cache file into memory (this object)
			parseCache(cacheDocument);
			// Update cache against files in the wizard package directory
			updateCache();
		} else {
			//Generate the cache file from scratch
			if (debug) {
				System.out.println("Creating new cache file at: " + cacheLocation);
			}

			buildCache();
			saveCache();
		}
		
	}

	/**
	 * Add all wizard packages to this WizardCache LinkedList
	 */
	private void buildCache() {
		// Get a array of the wizard package Files
		File[] WizardFiles = WizPackageHandler.getWizardFiles();
		
		// For each file in the array, add it to this WizardCache LinkedList
		for(int i =0; i < WizardFiles.length; ++i) {
			try {
				add(WizPackageHandler.CacheItemFromPackage(WizardFiles[i]));
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Save the cache list of wizards to an XML file
	 */
	public void saveCache() {
		if (debug) {
			System.out.println("WizardCache.saveCache() started");
		}

		// Clear cacheDocument, and create from scratch  
		cacheDocument = null;
		try	{
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			DocumentBuilder db = dbf.newDocumentBuilder();
			cacheDocument = db.newDocument();
		}
		catch (Exception e ) {
			cacheDocument = null;
			System.out.println("exception" + e);
		}
		
		// Set up basic XML file contents 
		Element element = cacheDocument.createElement("wizards");
		cacheDocument.appendChild(element);
		Element root = cacheDocument.getDocumentElement();
		
		// Convert LinkedList cache to an DOM Document
		for (int i = 0; i < size(); i++) {
			//Get values
			String versionVal = ((WizardCacheItem)get(i)).getVersion();
			String fileVal = ((WizardCacheItem)get(i)).getFile();
			String titleVal = ((WizardCacheItem)get(i)).getTitle();
			String descriptionVal = ((WizardCacheItem)get(i)).getDescription();
			
			if (debug) {
				System.out.println("");
				System.out.println("versionVal: " + versionVal);
				System.out.println("fileVal: " + fileVal);
				System.out.println("titleVal: " + titleVal);
				System.out.println("descriptionVal: " + descriptionVal);
			}
			
			//Construct a "wizard" element for this cache entry, with attributes
			Element wizardEl = cacheDocument.createElement("wizard");
			wizardEl.setAttribute("version", versionVal);
			wizardEl.setAttribute("file", fileVal);
			
			//Add the title child element
			Element titleEl = cacheDocument.createElement("title");
			titleEl.setTextContent(titleVal);
			wizardEl.appendChild(titleEl);
			
			//Add the description child element
			Element descriptionEl = cacheDocument.createElement("description");
			descriptionEl.setTextContent(descriptionVal);
			wizardEl.appendChild(descriptionEl);

			//Add this wizard element to the Document
			root.appendChild(wizardEl);
		}
		
		// Save DOM Document
		writeXmlFile(cacheDocument, wizardCacheFile);
	}
	
	/** Parse the cache file (XML)
	 * Convert "in-memory" document to a LinkedList
	 */

	private void parseCache(Document cacheDoc){
		
		NodeList wizardCacheItems = cacheDoc.getElementsByTagName("wizard");
		if (debug) {
			System.out.println("~~~~~~~~~~~ Started parseCache() ~~~~~~~~~~~~~");
			System.out.println("Number of wizardCacheItems in the cacheDoc: " + wizardCacheItems.getLength() );
		}
		
		//Setup XPath stuff
		XPathFactory xpf = XPathFactory.newInstance();
		XPath xpath= xpf.newXPath();
		
		//Add each wizard in the cache file to a LinkedList 
		for(int i=0; i < wizardCacheItems.getLength(); ++i) {
			String title = null;
			String description = null;
			String file = null;
			String version = null;
			
			if (debug) {
				System.out.println("wizardCacheItems.item(i): " + wizardCacheItems.item(i) );
			}
			
			try {
				title = xpath.evaluate("title", wizardCacheItems.item(i));
				file = xpath.evaluate("@file", wizardCacheItems.item(i));
				version = xpath.evaluate("@version", wizardCacheItems.item(i));
				description = xpath.evaluate("description", wizardCacheItems.item(i));
				
			} catch (XPathExpressionException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			if (debug) {
				System.out.println("");
				System.out.println("Title:       '" + title + "'");
				System.out.println("Description: '" + description + "'");
				System.out.println("File:        '" + file + "'");
				System.out.println("Version:     '" + version + "'");
			}
			
			// Add details as a item in the list 
			add( new WizardCacheItem(title, description, file, version));
		}
		
	}

	/**
	 * Update cache file from wizard package files
	 * @param cacheLocation file name
	 */
	public void updateCache(){
		if (debug) {
			System.out.println("");
			System.out.println("WizardCache.updateCache() started");
		}
		boolean cacheUpdated = false;
		
		/*
		 *  Compare the cache file LinkedList to the wizard directory contents
		 */
		
		File[] WizardFiles = WizPackageHandler.getWizardFiles();
		
		//For each file in the array,
		for(int i =0; i < WizardFiles.length; ++i) {
			boolean packageCached = false;
			
			// Get the file name
			String fileName = WizardFiles[i].getName();
			if (debug) {
				System.out.println("");
				System.out.println("Wizard file: " + fileName );
			}
			
			//For each cache entry
			for(int j =0; j < size(); ++j){
				// Get the cached file name
				String cacheFileName = ((WizardCacheItem)get(j)).getFile();
				if (debug) {
					System.out.println("Wizard filename found in cache: " + cacheFileName );
				}
				
				// Check whether the filenames between the file list and the cache entry match
				if (fileName.equals(cacheFileName)) {
					// If no entry exists, mark that we want to add it
					packageCached = true;
					break;
				}
			}

			if (debug) {
				System.out.println("packageCached: " + packageCached );
			}
			//If we need to, update the cache, and mark the cache as updated
			if (!packageCached){
				try {
					// update the cache
					add(WizPackageHandler.CacheItemFromPackage(WizardFiles[i]));
				} catch (IOException e) {
					// Auto-generated catch block
					e.printStackTrace();
				}
				cacheUpdated = true;
			}
		}
		/*
		 * TODO: Check for packages listed in the cache but are no longer in 
		 * the package directory, and remove them from the cache.
		 */
		
		
		// If changed, save the cache to an XML file
		if (cacheUpdated){
			if (debug) {
				System.out.println("Cache updated, so save it.");
			}
			saveCache();
		};
		
	}
	
	//	 This method writes a DOM document to a file
	// TODO: put this in a sensible XML utility class.
	public static void writeXmlFile(Document doc, String filename) {
		// Prepare the output file
		File file = new File(filename);
		
		writeXmlFile(doc, file);
	}
	
	//	 This method writes a DOM document to a file
	public static void writeXmlFile(Document doc, File file) {
		try {
			// Prepare the DOM document for writing
			Source source = new DOMSource(doc);
			
			// Prepare the output file
			Result result = new StreamResult(file);
			
			// Write the DOM document to the file
			Transformer xformer = TransformerFactory.newInstance().newTransformer();
			xformer.transform(source, result);
		} catch (TransformerConfigurationException e) {
		} catch (TransformerException e) {
		}
	}
}
